/*
Module : IniObj.cpp
Purpose: Defines the implementation for an ATL COM object to support reading / writing to INI files
Created: PJN / 25-11-2001
History: PJN / 20-03-2003 1. Fixed a bug in ReadString and ReadKeys where the default value is set to an empty string
                          and the value(s) is not present in the ini file. Thanks to Serhiy Pavlov for reporting 
                          this problem.
         PJN / 20-03-2003 1. Updated all the functions which return COM SAFEARRAY's to instead return SAFEARRAY's of 
                          VARIANT's. This avoids problems which JScript has with VB style arrays.
                          2. Updated code which returns COM errors to make use of MAKE_HRESULT.
         PJN / 25-08-2003 1. Code now does not return an error if get_Sections detects no sections in the ini file
                          2. Code now does not return an error if ReadSection detects no key value pairs in a section
         PJN / 18-07-2006 1. Updated copyright details.
                          2. Updated documentation to use the same style as the web site.  
                          3. Code now uses SEVERITY_ERROR define for MAKE_HRESULT calls.
                          4. Updated code to clean compile on VC 2005
                          5. All calls to SysAllocString are now checked for failure.
                          6. All C++ heap allocations via "new" are now checked for failure.
         PJN / 30-10-2016 1. Updated copyright details.
                          2. Made IniUtil freeware instead of shareware. Now included in the zip file download is the
                          source code for IniUtil
                          3. Updated code to clean compile on VC 2010 - 2015.
                          4. The binaries include in the download are now compiled using VC 2015 and include x64 
                          binaries as well as x86 binaries.
                          5. Replaced NULL with nullptr throughout the codebase
                          6. Removed Non-Unicode build configurations from the project.
         PJN / 02-05-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly using /permissive-.
         PJN / 05-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that 
                          the code will now only compile on VC 2017 or later.
         PJN / 10-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 07-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2001 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

All rights reserved.

*/

//////////////////// Includes /////////////////////////////////////////////////

#include "StdAfx.h"
#include "IniUtil_i.h"
#include "IniObj.h"


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::get_Filename(BSTR* pbstrFilename) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pbstrFilename == nullptr)
    return Error(L"Filename is a null pointer", IID_IIniUtil, E_POINTER);

  //Return a copy of our string
  return m_sFilename.CopyTo(pbstrFilename);
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::put_Filename(BSTR bstrFilename) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (bstrFilename == nullptr)
    return Error(L"Filename is a null pointer", IID_IIniUtil, E_POINTER);

  //Just hive away a value into our member variable
  return m_sFilename.AssignBSTR(bstrFilename);
}

HRESULT CIniObj::MakeSafeArrayVariantFromMultiSZString(_In_z_ LPCWSTR pszString, _Out_ VARIANT* pvVariant)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(pvVariant != nullptr);

  //First work out how many strings there are in the multi sz string
  int nStrings{0};
#pragma warning(suppress: 26429)
  LPCWSTR pszCurrentString{pszString};
#pragma warning(suppress: 26481)
  while (pszCurrentString[0] != L'\0')
  {
    ++nStrings;
#pragma warning(suppress: 26481)
    pszCurrentString += (wcslen(pszCurrentString) + 1);
  }

  //Set the attributes in the VARIANT
  VariantInit(pvVariant);
  pvVariant->vt = VT_ARRAY | VT_VARIANT; //an array of variant strings

  //Create the safe array
  SAFEARRAYBOUND bound;
  bound.cElements = nStrings;
  bound.lLbound = 1;
  SAFEARRAY* pSafeArray{SafeArrayCreate(VT_VARIANT, 1, &bound)};
  if (pSafeArray == nullptr)
    return Error(L"Error calling SafeArrayCreate", IID_IIniUtil, E_OUTOFMEMORY);

  //Set each array element's value
  pszCurrentString = pszString;
  for (long i{1}; i<=nStrings; i++)
  {
    VARIANT varElement;
    varElement.vt = VT_BSTR;
    varElement.bstrVal = SysAllocString(pszCurrentString);
    if (varElement.bstrVal == nullptr)
    {
      VariantClear(pvVariant);
      return Error(L"Error calling SysAllocString", IID_IIniUtil, E_OUTOFMEMORY);
    }
#pragma warning(suppress: 26486)
    const HRESULT hr{SafeArrayPutElement(pSafeArray, &i, &varElement)};
    if (FAILED(hr))
      return Error(L"Error calling SafeArrayPutElement", IID_IIniUtil, hr);
#pragma warning(suppress: 26481)
    pszCurrentString += (wcslen(pszCurrentString) + 1);
  }

  //Assign the safearray into the VARIANT before we return
  pvVariant->parray = pSafeArray;
  return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::get_Sections(VARIANT* pvSections) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pvSections == nullptr)
    return Error(L"Sections is a null pointer", IID_IIniUtil, E_POINTER);

  DWORD dwSize{1024};
  BOOL bSuccess{FALSE};
  HRESULT hr{S_OK};
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    ATL::CHeapPtr<wchar_t> szSections;
    if (!szSections.Allocate(dwSize))
      return Error(L"Error allocating memory in Sections", IID_IIniUtil, E_OUTOFMEMORY);

    //Call the SDK function
    const DWORD dwRetrieved{GetPrivateProfileSectionNamesW(szSections, dwSize, m_sFilename)};
    if (dwRetrieved == (dwSize - 2))
    {
      //Realloc the array by doubling its size ready for the next loop around
      dwSize *= 2;
    }
    else if (dwRetrieved == 0)
    {
      const DWORD dwError{GetLastError()};

      //Only consider it an error if no data was retrieved and GetLastError flagged an error
      if (dwError != ERROR_SUCCESS)
        return Error(L"Error calling GetPrivateProfileSectionNamesW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, dwError));
      else
      {
        bSuccess = TRUE;
        hr = MakeSafeArrayVariantFromMultiSZString(L"\0\0", pvSections);
      }
    }
    else
    {
      //Convert from a multi sz string to a VARIANT safearray
      bSuccess = TRUE;
      hr = MakeSafeArrayVariantFromMultiSZString(szSections, pvSections);
    }
  }

  return hr;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::WriteString(BSTR bstrSection, BSTR bstrKey, BSTR bstrValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Call the SDK function
  const BOOL bSuccess{WritePrivateProfileStringW(bstrSection, bstrKey, bstrValue, m_sFilename)};

  //Handle the failure of the API
  if (!bSuccess)
    return Error(L"Error calling WritePrivateProfileStringW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
  else
    return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::WriteInteger(BSTR bstrSection, BSTR bstrKey, int nValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  ATL::CStringW sNumber;
  sNumber.Format(L"%d", nValue);

  //Call the SDK function
  const BOOL bSuccess{WritePrivateProfileStringW(bstrSection, bstrKey, sNumber, m_sFilename)};

  //Handle the failure of the API
  if (!bSuccess)
    return Error(L"Error calling WritePrivateProfileStringW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
  else
    return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::WriteVariant(BSTR bstrSection, BSTR bstrKey, VARIANT vValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if ((vValue.vt & VT_UNKNOWN) == VT_UNKNOWN)
    return Error(L"Writing of COM interfaces to ini files is not supported", IID_IIniUtil, E_INVALIDARG);
  if ((vValue.vt & VT_DISPATCH) == VT_DISPATCH)
    return Error(L"Writing of COM dispatch interfaces to ini files is not supported", IID_IIniUtil, E_INVALIDARG);
  if ((vValue.vt & VT_BYREF) == VT_BYREF)
    return Error(L"Writing of variants passed by reference to ini files is not supported", IID_IIniUtil, E_INVALIDARG);
  if ((vValue.vt & VT_ARRAY) == VT_ARRAY)
    return Error(L"Writing of arrays to ini files is not supported", IID_IIniUtil, E_INVALIDARG);

  //Call the SDK function
  const BOOL bSuccess{WritePrivateProfileStructW(bstrSection, bstrKey, &vValue, sizeof(VARIANT), m_sFilename)};

  //Handle the failure of the API
  if (!bSuccess)
    return Error(L"Error calling WritePrivateProfileStructW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
  else
    return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::WriteSection(BSTR bstrSection, VARIANT vKeyValues) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);

  // checking if it is a one-dimensional array
  if (vKeyValues.parray->cDims != 1)
    return Error(L"KeyValues is not a 1 dimensional array", IID_IIniUtil, E_INVALIDARG);

  //Get the array bounds
  long lLBound{0};
  HRESULT hr{SafeArrayGetLBound(vKeyValues.parray, 1, &lLBound)};
  if (FAILED(hr))
    return Error(L"Error calling SafeArrayGetLBound", IID_IIniUtil, hr);

  long lUBound{0};
  hr = SafeArrayGetUBound(vKeyValues.parray, 1, &lUBound);
  if (FAILED(hr))
    return Error(L"Error calling SafeArrayGetUBound", IID_IIniUtil, hr);

  if (vKeyValues.vt == (VT_ARRAY|VT_BSTR)) //an array of strings
  {
    //Work out the size of the doubly null terminated string we need
    size_t nStringSize{0};
    for (long i{lLBound}; i<=lUBound; i++)
    {
      BSTR bstrKeyValue{nullptr};
      hr = SafeArrayGetElement(vKeyValues.parray, &i, &bstrKeyValue);
      if (FAILED(hr))
        return Error(L"Error calling SafeArrayGetElement", IID_IIniUtil, hr);

      nStringSize += SysStringLen(bstrKeyValue);
      nStringSize++; //Include the null terminator
    }
    nStringSize++; //Include the final null terminator

    //Allocate the actual memory for the string
    ATL::CHeapPtr<wchar_t> szKeyValues;
    if (!szKeyValues.Allocate(nStringSize))
      return Error(L"Error allocating memory", IID_IIniUtil, E_OUTOFMEMORY);

    //Now actually form the string
    int nStringOffset{0};
    for (long i{lLBound}; i<=lUBound; i++)
    {
      BSTR bstrKeyValue{nullptr};
      hr = SafeArrayGetElement(vKeyValues.parray, &i, &bstrKeyValue);
      if (FAILED(hr))
        return Error(L"Error calling SafeArrayGetElement", IID_IIniUtil, hr);

      const UINT nCurrentSize{SysStringLen(bstrKeyValue)};
#pragma warning(suppress: 26481)
      wcscpy_s(szKeyValues + nStringOffset, nStringSize - nStringOffset, bstrKeyValue);
      nStringOffset += (nCurrentSize+1);
    }
#pragma warning(suppress: 26481)
    szKeyValues[nStringOffset] = L'\0'; //Include the final null terminator

    //Call the SDK function
    const BOOL bSuccess{WritePrivateProfileSectionW(bstrSection, szKeyValues, m_sFilename)};
    if (!bSuccess)
      hr = Error(L"Error calling WritePrivateProfileSectionW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
  }
  else if (vKeyValues.vt == (VT_ARRAY|VT_VARIANT)) //an array of variants
  {
    //Work out the size of the doubly null terminated string we need
    size_t nStringSize{0};
    for (long i{lLBound}; i<=lUBound; i++)
    {
      VARIANT varKeyValue;
      hr = SafeArrayGetElement(vKeyValues.parray, &i, &varKeyValue);
      if (FAILED(hr))
        return Error(L"Error calling SafeArrayGetElement", IID_IIniUtil, hr);
      if (varKeyValue.vt != VT_BSTR)
        return Error(L"KeyValues is not an array of variant strings", IID_IIniUtil, E_INVALIDARG);

#pragma warning(suppress: 26486)
      nStringSize += SysStringLen(varKeyValue.bstrVal);
      nStringSize++; //Include the null terminator
    }
    nStringSize++; //Include the final null terminator

    //Allocate the actual memory for the string
    ATL::CHeapPtr<wchar_t> szKeyValues;
    if (!szKeyValues.Allocate(nStringSize))
      return Error(L"Error allocating memory", IID_IIniUtil, E_OUTOFMEMORY);

    //Now actually form the string
    int nStringOffset{0};
    for (long i{lLBound}; i<=lUBound; i++)
    {
      VARIANT varKeyValue;
      hr = SafeArrayGetElement(vKeyValues.parray, &i, &varKeyValue);
      if (FAILED(hr))
        return Error(L"Error calling SafeArrayGetElement", IID_IIniUtil, hr);

#pragma warning(suppress: 26486)
      const UINT nCurrentSize{SysStringLen(varKeyValue.bstrVal)};
#pragma warning(suppress: 26481)
      wcscpy_s(szKeyValues + nStringOffset, nStringSize - nStringOffset, varKeyValue.bstrVal);
      nStringOffset += (nCurrentSize+1);
    }
#pragma warning(suppress: 26481)
    szKeyValues[nStringOffset] = L'\0'; //Include the final null terminator

    //Call the SDK function
    const BOOL bSuccess{WritePrivateProfileSectionW(bstrSection, szKeyValues, m_sFilename)};
    if (!bSuccess)
      hr = Error(L"Error calling WritePrivateProfileSectionW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
  }
  else
    return Error(L"KeyValues is not an array of strings or variant strings", IID_IIniUtil, E_INVALIDARG);
  return hr;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::ReadString(BSTR bstrSection, BSTR bstrKey, BSTR bstrDefaultValue, BSTR* pbstrValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pbstrValue == nullptr)
    return Error(L"Value is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrKey == nullptr)
    return Error(L"Key is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrDefaultValue == nullptr)
    return Error(L"DefaultValue is a null pointer", IID_IIniUtil, E_POINTER);

  DWORD dwSize{1024};
  BOOL bSuccess{FALSE};
  HRESULT hr{S_OK};
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    ATL::CHeapPtr<wchar_t> szValue;
    if (!szValue.Allocate(dwSize))
      return Error(L"Error allocating memory", IID_IIniUtil, E_OUTOFMEMORY);

    //Call the SDK function
    const DWORD dwRetrieved{GetPrivateProfileStringW(bstrSection, bstrKey, bstrDefaultValue, szValue, dwSize, m_sFilename)};
    if (dwRetrieved == (dwSize - 1))
      dwSize *= 2;
    else if ((dwRetrieved == 0) && wcslen(bstrDefaultValue))
      return Error(L"Error calling GetPrivateProfileStringW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));
    else
    {
      //Allocate the BSTR out parameter
      bSuccess = TRUE;
      *pbstrValue = SysAllocString(szValue);
      if (*pbstrValue == nullptr)
        return Error(L"Error allocating memory", IID_IIniUtil, E_OUTOFMEMORY);
      hr = S_OK;
    }
  }
  return hr;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::ReadInteger(BSTR bstrSection, BSTR bstrKey, int nDefaultValue, int* pnValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pnValue == nullptr)
    return Error(L"Value is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrKey == nullptr)
    return Error(L"Key is a null pointer", IID_IIniUtil, E_POINTER);

  //Call the SDK function
  *pnValue = GetPrivateProfileIntW(bstrSection, bstrKey, nDefaultValue, m_sFilename);

  return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::ReadVariant(BSTR bstrSection, BSTR bstrKey, VARIANT* pvValue) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pvValue == nullptr)
    return Error(L"Value is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrKey == nullptr)
    return Error(L"Key is a null pointer", IID_IIniUtil, E_POINTER);

  //Call the SDK function
  const BOOL bSuccess{GetPrivateProfileStructW(bstrSection, bstrKey, pvValue, sizeof(VARIANT), m_sFilename)};
  if (!bSuccess)
    return Error(L"Error calling GetPrivateProfileStructW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));

  return S_OK;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::ReadSection(BSTR bstrSection, VARIANT* pvKeyValues) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pvKeyValues == nullptr)
    return Error(L"KeyValues is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);

  DWORD dwSize{1024};
  BOOL bSuccess{FALSE};
  HRESULT hr{S_OK};
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    ATL::CHeapPtr<wchar_t> szKeyValues;
    if (!szKeyValues.Allocate(dwSize))
      return Error(L"Error allocating memory in ReadSection", IID_IIniUtil, E_OUTOFMEMORY);

    //Call the SDK function
    const DWORD dwRetrieved{GetPrivateProfileSectionW(bstrSection, szKeyValues, dwSize, m_sFilename)};
    if (dwRetrieved == (dwSize - 2))
    {
      //Realloc the array by doubling its size ready for the next loop around
      dwSize *= 2;
    }
    else if (dwRetrieved == 0)
    {
      const DWORD dwError{GetLastError()};

      //Only consider it an error if no data was retrieved and GetLastError flagged an error
      if (dwError != ERROR_SUCCESS)
        return Error(L"Error calling GetPrivateProfileSectionW", IID_IIniUtil, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, dwError));
      else
      {
        bSuccess = TRUE;
        hr = MakeSafeArrayVariantFromMultiSZString(L"\0\0", pvKeyValues);
      }
    }
    else
    {
      //Convert from a multi sz string to a VARIANT safearray
      bSuccess = TRUE;
      hr = MakeSafeArrayVariantFromMultiSZString(szKeyValues, pvKeyValues);
    }
  }

  return hr;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::ReadKeys(BSTR bstrSection, VARIANT* pvKeys) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pvKeys == nullptr)
    return Error(L"Keys is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);

  DWORD dwSize{1024};
  BOOL bSuccess{FALSE};
  HRESULT hr{S_OK};
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    ATL::CHeapPtr<wchar_t> szKeys;
    if (!szKeys.Allocate(dwSize))
      return Error(L"Error allocating memory", IID_IIniUtil, E_OUTOFMEMORY);
#pragma warning(suppress: 26481)
    szKeys[0] = L'\0';

    //Call the SDK function
    const DWORD dwRetrieved{GetPrivateProfileStringW(bstrSection, nullptr, L"", szKeys, dwSize, m_sFilename)};
    if (dwRetrieved == (dwSize - 2))
    {
      //Realloc the array by doubling its size ready for the next loop around
      dwSize *= 2;
    }
    else
    {
      //Convert from a multi sz string to a VARIANT safearray
      bSuccess = TRUE;
      hr = MakeSafeArrayVariantFromMultiSZString(szKeys, pvKeys);
    }
  }

  return hr;
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::DeleteSection(BSTR bstrSection) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);

  return WriteString(bstrSection, nullptr, nullptr);
}

#pragma warning(suppress: 26433)
STDMETHODIMP CIniObj::DeleteKey(BSTR bstrSection, BSTR bstrKey) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (bstrSection == nullptr)
    return Error(L"Section is a null pointer", IID_IIniUtil, E_POINTER);
  if (bstrKey == nullptr)
    return Error(L"Key is a null pointer", IID_IIniUtil, E_POINTER);

  return WriteString(bstrSection, bstrKey, nullptr);
}
